@@ -1,6 +1,6 @@
 /*
     SDL_ttf:  A companion library to SDL for working with TrueType (tm) fonts
-    Copyright (C) 1997-2004 Sam Lantinga
+    Copyright (C) 1997-2009 Sam Lantinga
 
     This library is free software; you can redistribute it and/or
     modify it under the terms of the GNU Library General Public
@@ -20,7 +20,7 @@
     slouken@libsdl.org
 */
 
-/* $Id: SDL_ttf.c 3284 2007-07-15 06:18:29Z slouken $ */
+/* $Id: SDL_ttf.c 5141 2009-10-18 20:47:04Z slouken $ */
 
 #include <math.h>
 #include <stdio.h>
@@ -42,6 +42,8 @@
 #include <ft2build.h>
 #include FT_FREETYPE_H
 #include FT_OUTLINE_H
+#include FT_STROKER_H
+#include FT_GLYPH_H
 #include FT_TRUETYPE_IDS_H
 
 #include "SDL.h"
@@ -88,7 +90,12 @@ struct _TTF_Font {
 	int lineskip;
 
 	/* The font style */
+	int face_style;
 	int style;
+	int outline;
+
+	/* Whether kerning is desired */
+	int kerning;
 
 	/* Extra width in glyph bounds for text styles */
 	int glyph_overhang;
@@ -110,8 +117,22 @@ struct _TTF_Font {
 
 	/* For non-scalable formats, we must remember which font index size */
 	int font_size_family;
+	
+	/* really just flags passed into FT_Load_Glyph */
+	int hinting;
 };
 
+/* Handle a style only if the font does not already handle it */
+#define TTF_HANDLE_STYLE_BOLD(font) (((font)->style & TTF_STYLE_BOLD) && \
+                                    !((font)->face_style & TTF_STYLE_BOLD))
+#define TTF_HANDLE_STYLE_ITALIC(font) (((font)->style & TTF_STYLE_ITALIC) && \
+                                      !((font)->face_style & TTF_STYLE_ITALIC))
+#define TTF_HANDLE_STYLE_UNDERLINE(font) ((font)->style & TTF_STYLE_UNDERLINE)
+#define TTF_HANDLE_STYLE_STRIKETHROUGH(font) ((font)->style & TTF_STYLE_STRIKETHROUGH)
+
+/* Font styles that does not impact glyph drawing */
+#define TTF_STYLE_NO_GLYPH_CHANGE	(TTF_STYLE_UNDERLINE | TTF_STYLE_STRIKETHROUGH)
+
 /* The FreeType font engine/library */
 static FT_Library library;
 static int TTF_initialized = 0;
@@ -145,6 +166,140 @@ static __inline__ void UNICODE_strcpy(Uint16 *dst, const Uint16 *src, int swap)
 	}
 }
 
+/* Gets the top row of the underline. The outline
+   is taken into account.
+*/
+static __inline__ int TTF_underline_top_row(TTF_Font *font)
+{
+	/* With outline, the underline_offset is underline_offset+outline. */
+	/* So, we don't have to remove the top part of the outline height. */
+	return font->ascent - font->underline_offset - 1;
+}
+
+/* Gets the bottom row of the underline. The outline
+   is taken into account.
+*/
+static __inline__ int TTF_underline_bottom_row(TTF_Font *font)
+{
+	int row = TTF_underline_top_row(font) + font->underline_height;
+	if( font->outline  > 0 ) {
+		/* Add underline_offset outline offset and */
+		/* the bottom part of the outline. */
+		row += font->outline * 2;
+	}
+	return row;
+}
+
+/* Gets the top row of the strikethrough. The outline
+   is taken into account.
+*/
+static __inline__ int TTF_strikethrough_top_row(TTF_Font *font)
+{
+	/* With outline, the first text row is 'outline'. */
+	/* So, we don't have to remove the top part of the outline height. */
+	return font->height / 2;
+}
+
+/* Gets the bottom row of the strikethrough. The outline
+   is taken into account.
+*/
+static __inline__ int TTF_strikethrough_bottom_row(TTF_Font *font)
+{
+	int row = TTF_strikethrough_top_row(font) + font->underline_height;
+	if( font->outline  > 0 ) {
+		/* Add first text row outline offset and */
+		/* the bottom part of the outline. */
+		row += font->outline * 2;
+	}
+	return row;
+}
+
+static void TTF_initLineMectrics(const TTF_Font *font, const SDL_Surface *textbuf, const int row, Uint8 **pdst, int *pheight)
+{
+	Uint8 *dst;
+	int height;
+
+	dst = (Uint8 *)textbuf->pixels;
+	if( row > 0 ) {
+		dst += row * textbuf->pitch;
+	}
+
+	height = font->underline_height;
+	/* Take outline into account */
+	if( font->outline > 0 ) {
+		height += font->outline * 2;
+	}
+	*pdst = dst;
+	*pheight = height;
+}
+
+/* Draw a solid line of underline_height (+ optional outline)
+   at the given row. The row value must take the
+   outline into account.
+*/
+static void TTF_drawLine_Solid(const TTF_Font *font, const SDL_Surface *textbuf, const int row)
+{
+	int line;
+	Uint8 *dst_check = (Uint8*)textbuf->pixels + textbuf->pitch * textbuf->h;
+	Uint8 *dst;
+	int height;
+
+	TTF_initLineMectrics(font, textbuf, row, &dst, &height);
+
+	/* Draw line */
+	for ( line=height; line>0 && dst < dst_check; --line ) {
+		/* 1 because 0 is the bg color */
+		memset( dst, 1, textbuf->w );
+		dst += textbuf->pitch;
+	}
+}
+
+/* Draw a shaded line of underline_height (+ optional outline)
+   at the given row. The row value must take the
+   outline into account.
+*/
+static void TTF_drawLine_Shaded(const TTF_Font *font, const SDL_Surface *textbuf, const int row)
+{
+	int line;
+	Uint8 *dst_check = (Uint8*)textbuf->pixels + textbuf->pitch * textbuf->h;
+	Uint8 *dst;
+	int height;
+
+	TTF_initLineMectrics(font, textbuf, row, &dst, &height);
+
+	/* Draw line */
+	for ( line=height; line>0 && dst < dst_check; --line ) {
+		memset( dst, NUM_GRAYS - 1, textbuf->w );
+		dst += textbuf->pitch;
+	}
+}
+
+/* Draw a blended line of underline_height (+ optional outline)
+   at the given row. The row value must take the
+   outline into account.
+*/
+static void TTF_drawLine_Blended(const TTF_Font *font, const SDL_Surface *textbuf, const int row, const Uint32 color)
+{
+	int line;
+	Uint32 *dst_check = (Uint32*)textbuf->pixels + textbuf->pitch/4 * textbuf->h;
+	Uint8 *dst8; /* destination, byte version */
+	Uint32 *dst;
+	int height;
+	int col;
+	Uint32 pixel = color | 0xFF000000; /* Amask */
+
+	TTF_initLineMectrics(font, textbuf, row, &dst8, &height);
+	dst = (Uint32 *) dst8;
+
+	/* Draw line */
+	for ( line=height; line>0 && dst < dst_check; --line ) {
+		for ( col=0; col < textbuf->w; ++col ) {
+			dst[col] = pixel;
+		}
+		dst += textbuf->pitch/4;
+	}
+}
+
 /* rcg06192001 get linked library's version. */
 const SDL_version *TTF_Linked_Version(void)
 {
@@ -342,10 +497,22 @@ TTF_Font* TTF_OpenFontIndexRW( SDL_RWops *src, int freesrc, int ptsize, long ind
 		font->height, font->lineskip);
 	printf("\tunderline_offset = %d, underline_height = %d\n",
 		font->underline_offset, font->underline_height);
+	printf("\tunderline_top_row = %d, strikethrough_top_row = %d\n",
+		TTF_underline_top_row(font), TTF_strikethrough_top_row(font));
 #endif
 
+	/* Initialize the font face style */
+	font->face_style = TTF_STYLE_NORMAL;
+	if ( font->face->style_flags & FT_STYLE_FLAG_BOLD ) {
+		font->face_style |= TTF_STYLE_BOLD;
+	}
+	if ( font->face->style_flags & FT_STYLE_FLAG_ITALIC ) {
+		font->face_style |= TTF_STYLE_ITALIC;
+	}
 	/* Set the default font style */
-	font->style = TTF_STYLE_NORMAL;
+	font->style = font->face_style;
+	font->outline = 0;
+	font->kerning = 1;
 	font->glyph_overhang = face->size->metrics.y_ppem / 10;
 	/* x offset = cos(((90.0-12)/360)*2*M_PI), or 12 degree angle */
 	font->glyph_italics = 0.207f;
@@ -424,7 +591,7 @@ static FT_Error Load_Glyph( TTF_Font* font, Uint16 ch, c_glyph* cached, int want
 	if ( ! cached->index ) {
 		cached->index = FT_Get_Char_Index( face, ch );
 	}
-	error = FT_Load_Glyph( face, cached->index, FT_LOAD_DEFAULT );
+	error = FT_Load_Glyph( face, cached->index, FT_LOAD_DEFAULT | font->hinting);
 	if( error ) {
 		return error;
 	}
@@ -460,10 +627,10 @@ static FT_Error Load_Glyph( TTF_Font* font, Uint16 ch, c_glyph* cached, int want
 		}
 		
 		/* Adjust for bold and italic text */
-		if( font->style & TTF_STYLE_BOLD ) {
+		if( TTF_HANDLE_STYLE_BOLD(font) ) {
 			cached->maxx += font->glyph_overhang;
 		}
-		if( font->style & TTF_STYLE_ITALIC ) {
+		if( TTF_HANDLE_STYLE_ITALIC(font) ) {
 			cached->maxx += (int)ceil(font->glyph_italics);
 		}
 		cached->stored |= CACHED_METRICS;
@@ -475,9 +642,10 @@ static FT_Error Load_Glyph( TTF_Font* font, Uint16 ch, c_glyph* cached, int want
 		int i;
 		FT_Bitmap* src;
 		FT_Bitmap* dst;
+		FT_Glyph bitmap_glyph = NULL;
 
 		/* Handle the italic style */
-		if( font->style & TTF_STYLE_ITALIC ) {
+		if( TTF_HANDLE_STYLE_ITALIC(font) ) {
 			FT_Matrix shear;
 
 			shear.xx = 1 << 16;
@@ -488,18 +656,33 @@ static FT_Error Load_Glyph( TTF_Font* font, Uint16 ch, c_glyph* cached, int want
 			FT_Outline_Transform( outline, &shear );
 		}
 
-		/* Render the glyph */
-		if ( mono ) {
-			error = FT_Render_Glyph( glyph, ft_render_mode_mono );
+		/* Render as outline */
+		if( (font->outline > 0) && glyph->format != FT_GLYPH_FORMAT_BITMAP ) {
+			FT_Stroker stroker;
+			FT_Get_Glyph( glyph, &bitmap_glyph );
+			error = FT_Stroker_New( library, &stroker );
+			if( error ) {
+				return error;
+			}
+			FT_Stroker_Set( stroker, font->outline * 64, FT_STROKER_LINECAP_ROUND, FT_STROKER_LINEJOIN_ROUND, 0 ); 
+			FT_Glyph_Stroke( &bitmap_glyph, stroker, 1 /* delete the original glyph */ );
+			FT_Stroker_Done( stroker );
+			/* Render the glyph */
+			error = FT_Glyph_To_Bitmap( &bitmap_glyph, mono ? ft_render_mode_mono : ft_render_mode_normal, 0, 1 );
+			if( error ) {
+				FT_Done_Glyph( bitmap_glyph );
+				return error;
+			}
+			src = &((FT_BitmapGlyph)bitmap_glyph)->bitmap;
 		} else {
-			error = FT_Render_Glyph( glyph, ft_render_mode_normal );
+			/* Render the glyph */
+			error = FT_Render_Glyph( glyph, mono ? ft_render_mode_mono : ft_render_mode_normal );
+			if( error ) {
+				return error;
+			}
+			src = &glyph->bitmap;
 		}
-		if( error ) {
-			return error;
-		}
-
 		/* Copy over information to cache */
-		src = &glyph->bitmap;
 		if ( mono ) {
 			dst = &cached->bitmap;
 		} else {
@@ -517,21 +700,21 @@ static FT_Error Load_Glyph( TTF_Font* font, Uint16 ch, c_glyph* cached, int want
 		 * FT_Render_Glyph() canreturn two-color bitmap or 4/16/256-
 		 * color graymap according to the format of embedded bitmap/
 		 * graymap. */
-		if ( glyph->bitmap.pixel_mode == FT_PIXEL_MODE_MONO ) {
+		if ( src->pixel_mode == FT_PIXEL_MODE_MONO ) {
 			dst->pitch *= 8;
-		} else if ( glyph->bitmap.pixel_mode == FT_PIXEL_MODE_GRAY2 ) {
+		} else if ( src->pixel_mode == FT_PIXEL_MODE_GRAY2 ) {
 			dst->pitch *= 4;
-		} else if ( glyph->bitmap.pixel_mode == FT_PIXEL_MODE_GRAY4 ) {
+		} else if ( src->pixel_mode == FT_PIXEL_MODE_GRAY4 ) {
 			dst->pitch *= 2;
 		}
 
 		/* Adjust for bold and italic text */
-		if( font->style & TTF_STYLE_BOLD ) {
+		if( TTF_HANDLE_STYLE_BOLD(font) ) {
 			int bump = font->glyph_overhang;
 			dst->pitch += bump;
 			dst->width += bump;
 		}
-		if( font->style & TTF_STYLE_ITALIC ) {
+		if( TTF_HANDLE_STYLE_ITALIC(font) ) {
 			int bump = (int)ceil(font->glyph_italics);
 			dst->pitch += bump;
 			dst->width += bump;
@@ -551,7 +734,7 @@ static FT_Error Load_Glyph( TTF_Font* font, Uint16 ch, c_glyph* cached, int want
 					unsigned char *srcp = src->buffer + soffset;
 					unsigned char *dstp = dst->buffer + doffset;
 					int j;
-					if ( glyph->bitmap.pixel_mode == FT_PIXEL_MODE_MONO ) {
+					if ( src->pixel_mode == FT_PIXEL_MODE_MONO ) {
 						for ( j = 0; j < src->width; j += 8 ) {
 							unsigned char ch = *srcp++;
 							*dstp++ = (ch&0x80) >> 7;
@@ -570,7 +753,7 @@ static FT_Error Load_Glyph( TTF_Font* font, Uint16 ch, c_glyph* cached, int want
 							ch <<= 1;
 							*dstp++ = (ch&0x80) >> 7;
 						}
-					}  else if ( glyph->bitmap.pixel_mode == FT_PIXEL_MODE_GRAY2 ) {
+					}  else if ( src->pixel_mode == FT_PIXEL_MODE_GRAY2 ) {
 						for ( j = 0; j < src->width; j += 4 ) {
 							unsigned char ch = *srcp++;
 							*dstp++ = (((ch&0xA0) >> 6) >= 0x2) ? 1 : 0;
@@ -581,7 +764,7 @@ static FT_Error Load_Glyph( TTF_Font* font, Uint16 ch, c_glyph* cached, int want
 							ch <<= 2;
 							*dstp++ = (((ch&0xA0) >> 6) >= 0x2) ? 1 : 0;
 						}
-					} else if ( glyph->bitmap.pixel_mode == FT_PIXEL_MODE_GRAY4 ) {
+					} else if ( src->pixel_mode == FT_PIXEL_MODE_GRAY4 ) {
 						for ( j = 0; j < src->width; j += 2 ) {
 							unsigned char ch = *srcp++;
 							*dstp++ = (((ch&0xF0) >> 4) >= 0x8) ? 1 : 0;
@@ -594,7 +777,7 @@ static FT_Error Load_Glyph( TTF_Font* font, Uint16 ch, c_glyph* cached, int want
 							*dstp++ = (ch >= 0x80) ? 1 : 0;
 						}
 					}
-				} else if ( glyph->bitmap.pixel_mode == FT_PIXEL_MODE_MONO ) {
+				} else if ( src->pixel_mode == FT_PIXEL_MODE_MONO ) {
 					/* This special case wouldn't
 					 * be here if the FT_Render_Glyph()
 					 * function wasn't buggy when it tried
@@ -619,7 +802,7 @@ static FT_Error Load_Glyph( TTF_Font* font, Uint16 ch, c_glyph* cached, int want
 							ch <<= 1;
 						}
 					}
-				} else if ( glyph->bitmap.pixel_mode == FT_PIXEL_MODE_GRAY2 ) {
+				} else if ( src->pixel_mode == FT_PIXEL_MODE_GRAY2 ) {
 					unsigned char *srcp = src->buffer + soffset;
 					unsigned char *dstp = dst->buffer + doffset;
 					unsigned char ch;
@@ -635,7 +818,7 @@ static FT_Error Load_Glyph( TTF_Font* font, Uint16 ch, c_glyph* cached, int want
 							ch <<= 2;
 						}
 					}
-				} else if ( glyph->bitmap.pixel_mode == FT_PIXEL_MODE_GRAY4 ) {
+				} else if ( src->pixel_mode == FT_PIXEL_MODE_GRAY4 ) {
 					unsigned char *srcp = src->buffer + soffset;
 					unsigned char *dstp = dst->buffer + doffset;
 					unsigned char ch;
@@ -659,7 +842,7 @@ static FT_Error Load_Glyph( TTF_Font* font, Uint16 ch, c_glyph* cached, int want
 		}
 
 		/* Handle the bold style */
-		if ( font->style & TTF_STYLE_BOLD ) {
+		if ( TTF_HANDLE_STYLE_BOLD(font) ) {
 			int row;
 			int col;
 			int offset;
@@ -691,6 +874,11 @@ static FT_Error Load_Glyph( TTF_Font* font, Uint16 ch, c_glyph* cached, int want
 		} else {
 			cached->stored |= CACHED_PIXMAP;
 		}
+
+		/* Free outlined glyph */
+		if( bitmap_glyph ) {
+			FT_Done_Glyph( bitmap_glyph );
+		}
 	}
 
 	/* We're done, mark this glyph cached */
@@ -795,6 +983,16 @@ int TTF_FontLineSkip(const TTF_Font *font)
 	return(font->lineskip);
 }
 
+int TTF_GetFontKerning(const TTF_Font *font)
+{
+	return(font->kerning);
+}
+
+void TTF_SetFontKerning(TTF_Font *font, int allowed)
+{
+	font->kerning = allowed;
+}
+
 long TTF_FontFaces(const TTF_Font *font)
 {
 	return(font->face->num_faces);
@@ -815,6 +1013,11 @@ char *TTF_FontFaceStyleName(const TTF_Font *font)
 	return(font->face->style_name);
 }
 
+int TTF_GlyphIsProvided(const TTF_Font *font, Uint16 ch)
+{
+  return(FT_Get_Char_Index(font->face, ch));
+}
+
 int TTF_GlyphMetrics(TTF_Font *font, Uint16 ch,
                      int* minx, int* maxx, int* miny, int* maxy, int* advance)
 {
@@ -831,7 +1034,7 @@ int TTF_GlyphMetrics(TTF_Font *font, Uint16 ch,
 	}
 	if ( maxx ) {
 		*maxx = font->current->maxx;
-		if( font->style & TTF_STYLE_BOLD ) {
+		if( TTF_HANDLE_STYLE_BOLD(font) ) {
 			*maxx += font->glyph_overhang;
 		}
 	}
@@ -843,7 +1046,7 @@ int TTF_GlyphMetrics(TTF_Font *font, Uint16 ch,
 	}
 	if ( advance ) {
 		*advance = font->current->advance;
-		if( font->style & TTF_STYLE_BOLD ) {
+		if( TTF_HANDLE_STYLE_BOLD(font) ) {
 			*advance += font->glyph_overhang;
 		}
 	}
@@ -910,6 +1113,7 @@ int TTF_SizeUNICODE(TTF_Font *font, const Uint16 *text, int *w, int *h)
 	FT_Error error;
 	FT_Long use_kerning;
 	FT_UInt prev_index = 0;
+	int outline_delta = 0;
 
 	/* Initialize everything to 0 */
 	if ( ! TTF_initialized ) {
@@ -922,7 +1126,12 @@ int TTF_SizeUNICODE(TTF_Font *font, const Uint16 *text, int *w, int *h)
 	swapped = TTF_byteswapped;
 
 	/* check kerning */
-	use_kerning = FT_HAS_KERNING( font->face );
+	use_kerning = FT_HAS_KERNING( font->face ) && font->kerning;
+
+	/* Init outline handling */
+	if ( font->outline  > 0 ) {
+		outline_delta = font->outline * 2;
+	}
 
 	/* Load each character and sum it's bounding box */
 	x= 0;
@@ -983,7 +1192,7 @@ int TTF_SizeUNICODE(TTF_Font *font, const Uint16 *text, int *w, int *h)
 		if ( minx > z ) {
 			minx = z;
 		}
-		if ( font->style & TTF_STYLE_BOLD ) {
+		if ( TTF_HANDLE_STYLE_BOLD(font) ) {
 			x += font->glyph_overhang;
 		}
 		if ( glyph->advance > glyph->maxx ) {
@@ -1007,14 +1216,23 @@ int TTF_SizeUNICODE(TTF_Font *font, const Uint16 *text, int *w, int *h)
 
 	/* Fill the bounds rectangle */
 	if ( w ) {
-		*w = (maxx - minx);
+		/* Add outline extra width */
+		*w = (maxx - minx) + outline_delta;
 	}
 	if ( h ) {
-#if 0 /* This is correct, but breaks many applications */
-		*h = (maxy - miny);
-#else
-		*h = font->height;
-#endif
+		/* Some fonts descend below font height (FletcherGothicFLF) */
+		/* Add outline extra height */
+		*h = (font->ascent - miny) + outline_delta;
+		if ( *h < font->height ) {
+			*h = font->height;
+		}
+		/* Update height according to the needs of the underline style */
+		if( TTF_HANDLE_STYLE_UNDERLINE(font) ) {
+			int bottom_row = TTF_underline_bottom_row(font);
+			if ( *h < bottom_row ) {
+				*h = bottom_row;
+			}
+		}
 	}
 	return status;
 }
@@ -1095,11 +1313,10 @@ SDL_Surface *TTF_RenderUNICODE_Solid(TTF_Font *font,
 	FT_UInt prev_index = 0;
 
 	/* Get the dimensions of the text surface */
-	if( ( TTF_SizeUNICODE(font, text, &width, NULL) < 0 ) || !width ) {
+	if( ( TTF_SizeUNICODE(font, text, &width, &height) < 0 ) || !width ) {
 		TTF_SetError( "Text has zero width" );
 		return NULL;
 	}
-	height = font->height;
 
 	/* Create the target surface */
 	textbuf = SDL_AllocSurface(SDL_SWSURFACE, width, height, 8, 0, 0, 0, 0);
@@ -1122,7 +1339,7 @@ SDL_Surface *TTF_RenderUNICODE_Solid(TTF_Font *font,
 	SDL_SetColorKey( textbuf, SDL_SRCCOLORKEY, 0 );
 
 	/* check kerning */
-	use_kerning = FT_HAS_KERNING( font->face );
+	use_kerning = FT_HAS_KERNING( font->face ) && font->kerning;
 	
 	/* Load and render each character */
 	xstart = 0;
@@ -1157,7 +1374,7 @@ SDL_Surface *TTF_RenderUNICODE_Solid(TTF_Font *font,
 		/* Ensure the width of the pixmap is correct. On some cases,
 		 * freetype may report a larger pixmap than possible.*/
 		width = current->width;
-		if (width > glyph->maxx - glyph->minx) {
+		if (font->outline <= 0 && width > glyph->maxx - glyph->minx) {
 			width = glyph->maxx - glyph->minx;
 		}
 		/* do kerning, if possible AC-Patch */
@@ -1191,24 +1408,22 @@ SDL_Surface *TTF_RenderUNICODE_Solid(TTF_Font *font,
 		}
 
 		xstart += glyph->advance;
-		if ( font->style & TTF_STYLE_BOLD ) {
+		if ( TTF_HANDLE_STYLE_BOLD(font) ) {
 			xstart += font->glyph_overhang;
 		}
 		prev_index = glyph->index;
 	}
 
 	/* Handle the underline style */
-	if( font->style & TTF_STYLE_UNDERLINE ) {
-		row = font->ascent - font->underline_offset - 1;
-		if ( row >= textbuf->h) {
-			row = (textbuf->h-1) - font->underline_height;
-		}
-		dst = (Uint8 *)textbuf->pixels + row * textbuf->pitch;
-		for ( row=font->underline_height; row>0; --row ) {
-			/* 1 because 0 is the bg color */
-			memset( dst, 1, textbuf->w );
-			dst += textbuf->pitch;
-		}
+	if( TTF_HANDLE_STYLE_UNDERLINE(font) ) {
+		row = TTF_underline_top_row(font);
+		TTF_drawLine_Solid(font, textbuf, row);
+	}
+
+	/* Handle the strikethrough style */
+	if( TTF_HANDLE_STYLE_STRIKETHROUGH(font) ) {
+		row = TTF_strikethrough_top_row(font);
+		TTF_drawLine_Solid(font, textbuf, row);
 	}
 	return textbuf;
 }
@@ -1258,17 +1473,15 @@ SDL_Surface *TTF_RenderGlyph_Solid(TTF_Font *font, Uint16 ch, SDL_Color fg)
 	}
 
 	/* Handle the underline style */
-	if( font->style & TTF_STYLE_UNDERLINE ) {
-		row = font->ascent - font->underline_offset - 1;
-		if ( row >= textbuf->h) {
-			row = (textbuf->h-1) - font->underline_height;
-		}
-		dst = (Uint8 *)textbuf->pixels + row * textbuf->pitch;
-		for ( row=font->underline_height; row>0; --row ) {
-			/* 1 because 0 is the bg color */
-			memset( dst, 1, textbuf->w );
-			dst += textbuf->pitch;
-		}
+	if( TTF_HANDLE_STYLE_UNDERLINE(font) ) {
+		row = TTF_underline_top_row(font);
+		TTF_drawLine_Solid(font, textbuf, row);
+	}
+
+	/* Handle the strikethrough style */
+	if( TTF_HANDLE_STYLE_STRIKETHROUGH(font) ) {
+		row = TTF_strikethrough_top_row(font);
+		TTF_drawLine_Solid(font, textbuf, row);
 	}
 	return(textbuf);
 }
@@ -1355,11 +1568,10 @@ SDL_Surface* TTF_RenderUNICODE_Shaded( TTF_Font* font,
 	FT_UInt prev_index = 0;
 
 	/* Get the dimensions of the text surface */
-	if( ( TTF_SizeUNICODE(font, text, &width, NULL) < 0 ) || !width ) {
+	if( ( TTF_SizeUNICODE(font, text, &width, &height) < 0 ) || !width ) {
 		TTF_SetError("Text has zero width");
 		return NULL;
 	}
-	height = font->height;
 
 	/* Create the target surface */
 	textbuf = SDL_AllocSurface(SDL_SWSURFACE, width, height, 8, 0, 0, 0, 0);
@@ -1384,7 +1596,7 @@ SDL_Surface* TTF_RenderUNICODE_Shaded( TTF_Font* font,
 	}
 
 	/* check kerning */
-	use_kerning = FT_HAS_KERNING( font->face );
+	use_kerning = FT_HAS_KERNING( font->face ) && font->kerning;
 	
 	/* Load and render each character */
 	xstart = 0;
@@ -1418,7 +1630,7 @@ SDL_Surface* TTF_RenderUNICODE_Shaded( TTF_Font* font,
 		/* Ensure the width of the pixmap is correct. On some cases,
 		 * freetype may report a larger pixmap than possible.*/
 		width = glyph->pixmap.width;
-		if (width > glyph->maxx - glyph->minx) {
+		if (font->outline <= 0 && width > glyph->maxx - glyph->minx) {
 			width = glyph->maxx - glyph->minx;
 		}
 		/* do kerning, if possible AC-Patch */
@@ -1452,23 +1664,22 @@ SDL_Surface* TTF_RenderUNICODE_Shaded( TTF_Font* font,
 		}
 
 		xstart += glyph->advance;
-		if( font->style & TTF_STYLE_BOLD ) {
+		if( TTF_HANDLE_STYLE_BOLD(font) ) {
 			xstart += font->glyph_overhang;
 		}
 		prev_index = glyph->index;
 	}
 
 	/* Handle the underline style */
-	if( font->style & TTF_STYLE_UNDERLINE ) {
-		row = font->ascent - font->underline_offset - 1;
-		if ( row >= textbuf->h) {
-			row = (textbuf->h-1) - font->underline_height;
-		}
-		dst = (Uint8 *)textbuf->pixels + row * textbuf->pitch;
-		for ( row=font->underline_height; row>0; --row ) {
-			memset( dst, NUM_GRAYS - 1, textbuf->w );
-			dst += textbuf->pitch;
-		}
+	if( TTF_HANDLE_STYLE_UNDERLINE(font) ) {
+		row = TTF_underline_top_row(font);
+		TTF_drawLine_Shaded(font, textbuf, row);
+	}
+
+	/* Handle the strikethrough style */
+	if( TTF_HANDLE_STYLE_STRIKETHROUGH(font) ) {
+		row = TTF_strikethrough_top_row(font);
+		TTF_drawLine_Shaded(font, textbuf, row);
 	}
 	return textbuf;
 }
@@ -1527,16 +1738,15 @@ SDL_Surface* TTF_RenderGlyph_Shaded( TTF_Font* font,
 	}
 
 	/* Handle the underline style */
-	if( font->style & TTF_STYLE_UNDERLINE ) {
-		row = font->ascent - font->underline_offset - 1;
-		if ( row >= textbuf->h) {
-			row = (textbuf->h-1) - font->underline_height;
-		}
-		dst = (Uint8 *)textbuf->pixels + row * textbuf->pitch;
-		for ( row=font->underline_height; row>0; --row ) {
-			memset( dst, NUM_GRAYS - 1, textbuf->w );
-			dst += textbuf->pitch;
-		}
+	if( TTF_HANDLE_STYLE_UNDERLINE(font) ) {
+		row = TTF_underline_top_row(font);
+		TTF_drawLine_Shaded(font, textbuf, row);
+	}
+
+	/* Handle the strikethrough style */
+	if( TTF_HANDLE_STYLE_STRIKETHROUGH(font) ) {
+		row = TTF_strikethrough_top_row(font);
+		TTF_drawLine_Shaded(font, textbuf, row);
 	}
 	return textbuf;
 }
@@ -1615,12 +1825,12 @@ SDL_Surface *TTF_RenderUNICODE_Blended(TTF_Font *font,
 	FT_UInt prev_index = 0;
 
 	/* Get the dimensions of the text surface */
-	if ( (TTF_SizeUNICODE(font, text, &width, NULL) < 0) || !width ) {
+	if ( (TTF_SizeUNICODE(font, text, &width, &height) < 0) || !width ) {
 		TTF_SetError("Text has zero width");
 		return(NULL);
 	}
-	height = font->height;
 
+	/* Create the target surface */
 	textbuf = SDL_AllocSurface(SDL_SWSURFACE, width, height, 32,
 	                           0x00FF0000, 0x0000FF00, 0x000000FF, 0xFF000000);
 	if ( textbuf == NULL ) {
@@ -1632,7 +1842,7 @@ SDL_Surface *TTF_RenderUNICODE_Blended(TTF_Font *font,
 	dst_check = (Uint32*)textbuf->pixels + textbuf->pitch/4 * textbuf->h;
 
 	/* check kerning */
-	use_kerning = FT_HAS_KERNING( font->face );
+	use_kerning = FT_HAS_KERNING( font->face ) && font->kerning;
 	
 	/* Load and render each character */
 	xstart = 0;
@@ -1668,7 +1878,7 @@ SDL_Surface *TTF_RenderUNICODE_Blended(TTF_Font *font,
 		/* Ensure the width of the pixmap is correct. On some cases,
 		 * freetype may report a larger pixmap than possible.*/
 		width = glyph->pixmap.width;
-		if (width > glyph->maxx - glyph->minx) {
+		if (font->outline <= 0 && width > glyph->maxx - glyph->minx) {
 			width = glyph->maxx - glyph->minx;
 		}
 		/* do kerning, if possible AC-Patch */
@@ -1707,26 +1917,22 @@ SDL_Surface *TTF_RenderUNICODE_Blended(TTF_Font *font,
 		}
 
 		xstart += glyph->advance;
-		if ( font->style & TTF_STYLE_BOLD ) {
+		if ( TTF_HANDLE_STYLE_BOLD(font) ) {
 			xstart += font->glyph_overhang;
 		}
 		prev_index = glyph->index;
 	}
 
 	/* Handle the underline style */
-	if( font->style & TTF_STYLE_UNDERLINE ) {
-		row = font->ascent - font->underline_offset - 1;
-		if ( row >= textbuf->h) {
-			row = (textbuf->h-1) - font->underline_height;
-		}
-		dst = (Uint32 *)textbuf->pixels + row * textbuf->pitch/4;
-		pixel |= 0xFF000000; /* Amask */
-		for ( row=font->underline_height; row>0; --row ) {
-			for ( col=0; col < textbuf->w; ++col ) {
-				dst[col] = pixel;
-			}
-			dst += textbuf->pitch/4;
-		}
+	if( TTF_HANDLE_STYLE_UNDERLINE(font) ) {
+		row = TTF_underline_top_row(font);
+		TTF_drawLine_Blended(font, textbuf, row, pixel);
+	}
+
+	/* Handle the strikethrough style */
+	if( TTF_HANDLE_STYLE_STRIKETHROUGH(font) ) {
+		row = TTF_strikethrough_top_row(font);
+		TTF_drawLine_Blended(font, textbuf, row, pixel);
 	}
 	return(textbuf);
 }
@@ -1771,27 +1977,30 @@ SDL_Surface *TTF_RenderGlyph_Blended(TTF_Font *font, Uint16 ch, SDL_Color fg)
 	}
 
 	/* Handle the underline style */
-	if( font->style & TTF_STYLE_UNDERLINE ) {
-		row = font->ascent - font->underline_offset - 1;
-		if ( row >= textbuf->h) {
-			row = (textbuf->h-1) - font->underline_height;
-		}
-		dst = (Uint32 *)textbuf->pixels + row * textbuf->pitch/4;
-		pixel |= 0xFF000000; /* Amask */
-		for ( row=font->underline_height; row>0; --row ) {
-			for ( col=0; col < textbuf->w; ++col ) {
-				dst[col] = pixel;
-			}
-			dst += textbuf->pitch/4;
-		}
+	if( TTF_HANDLE_STYLE_UNDERLINE(font) ) {
+		row = TTF_underline_top_row(font);
+		TTF_drawLine_Blended(font, textbuf, row, pixel);
+	}
+
+	/* Handle the strikethrough style */
+	if( TTF_HANDLE_STYLE_STRIKETHROUGH(font) ) {
+		row = TTF_strikethrough_top_row(font);
+		TTF_drawLine_Blended(font, textbuf, row, pixel);
 	}
 	return(textbuf);
 }
 
 void TTF_SetFontStyle( TTF_Font* font, int style )
 {
-	font->style = style;
-	Flush_Cache( font );
+	int prev_style = font->style;
+	font->style = style | font->face_style;
+
+	/* Flush the cache if the style has changed.
+	 * Ignore UNDERLINE which does not impact glyph drawning.
+	 * */
+	if ( (font->style | TTF_STYLE_NO_GLYPH_CHANGE ) != ( prev_style | TTF_STYLE_NO_GLYPH_CHANGE )) {
+		Flush_Cache( font );
+	}
 }
 
 int TTF_GetFontStyle( const TTF_Font* font )
@@ -1799,6 +2008,35 @@ int TTF_GetFontStyle( const TTF_Font* font )
 	return font->style;
 }
 
+void TTF_SetFontOutline( TTF_Font* font, int outline )
+{
+	font->outline = outline;
+	Flush_Cache( font );
+}
+
+int TTF_GetFontOutline( const TTF_Font* font )
+{
+	return font->outline;
+}
+
+void TTF_SetFontHinting( TTF_Font* font, int hinting )
+{
+	if (hinting == TTF_HINTING_LIGHT)
+		font->hinting = FT_LOAD_TARGET_LIGHT;
+	else if (hinting == TTF_HINTING_MONO)
+		font->hinting = FT_LOAD_TARGET_MONO;
+	else if (hinting == TTF_HINTING_NONE)
+		font->hinting = FT_LOAD_NO_HINTING;
+	else
+		font->hinting = 0;
+	Flush_Cache( font );
+}
+
+int TTF_GetFontHinting( const TTF_Font* font )
+{
+	return font->hinting;
+}
+
 void TTF_Quit( void )
 {
 	if ( TTF_initialized ) {
